function model=POSO_PLS_perm(X,Y,options)
%options=POSO_PLS(X,Y)
%model=POSO_PLS(X,Y,options)
%
% options is a struct with fields:
%     -cvi: type of cross-validation. vector with indices or 'loo'/'r10'/r'20' (leave-one-out/random 10/random 20)
%     -autoselect: 0/1 automatic selection of components
%     -preprocX: cell array (length nBlocks) with preprocessing method for each X block. Either 'mean center' or 'autoscale'
%     -preprocY: preprocessing of Y. Either 'mean center' or 'autoscale'
%     -Amax: Vector (1 x nBlocks). Maximum number of components for each block.
%     -blockCombinations: cell array defining relevant combinations of blocks. default is all combinations.
%     -nCompsLocalPls: cell array (length blockCombinations) with number of components in local PLS models within each block combination
%     -nCompsForEachContribution: cell array (length blockCombinations) defining number of components from each blockCombination
%     
% model is a struct with fields:
%     -X: Input X data
%     -Y: Input Y data
%     -options: struct with fields described above. Not the same as input, the options are modified during modelling
%     -Scores: cell array with scores from each blockCombination
%     -Loadings: cell array with loadings from each blockCombination
%     -fittedY
%     -beta: coefficients from final prediction model (based on Scores)
%     -forprediction. Various parameters needed to predict new samples
%     -expVarX: cell array (blockCombinations x nBlocks) with cummulative explained X variances
%     -expVarY: cell array (blockCombinations x nY) with explained Y variances
%     -RMSEC: root mean squared error of calibration
%     -ANOVA: CV-anova results (need to run 'crossval_POSO_PLS')
%     -cvres: CV results (added after running 'crossval_POSO_PLS')
%
% Ingrid Mge
% Last modified: 12/8/2013



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% create options
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nargin<3
    
    
    model.cvi='r10';
    model.autoselect=0;
    model.preprocX='mean center';
    model.preprocY='mean center';
    model.Amax=[10];
    model.wb=1;
    model.blockCombinations=[];
    model.nCompsLocalPls=zeros(1,length(X)); % number of components in separate PLS models for each block
    model.nCompsForEachContribution={};
    if nargin>1
        for i=1:length(X)
            senteredX=X{i}.d-repmat(mean(X{i}.d),size(X{i}.d,1),1);
            model.Amax(i)=min([10 rank(senteredX)-1]);
        end
        dummy=repmat(model.preprocX,length(X),1);
        model.preprocX = mat2cell(dummy,ones(length(X),1),size(dummy,2));
        
        model.blockCombinations={};
        for i=1:(length(X)-1)
            r=combnk(1:length(X),length(X)-i+1);
            n=size(r,1);
            model.blockCombinations(end+1:end+n)=mat2cell(r,ones(n,1),length(X)-i+1)';
        end
        model.blockCombinations(end+1:end+length(X))=mat2cell(1:length(X),1,ones(length(X),1));
        model.nCompsLocalPls=zeros(1,length(X));
        model.nCompsForEachContribution=cell(1,length(model.blockCombinations));
    end
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % deal with input parameters
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    
else
    
    
    nBlocks=length(X);
    nObj=size(Y.d,1);
    nY=size(Y.d,2);
    
    if isfield(options,'blockCombinations')
        blockCombinations=options.blockCombinations;
    else
        blockCombinations={};
        for i=1:(nBlocks-1)
            r=combnk(1:nBlocks,nBlocks-i+1);
            n=size(r,1);
            blockCombinations(end+1:end+n)=mat2cell(r,ones(n,1),nBlocks-i+1)';
        end
        blockCombinations(end+1:end+nBlocks)=mat2cell(1:nBlocks,1,ones(nBlocks,1));
    end
    
    
    if isfield(options,'nCompsLocalPls')
        nCompsLocalPls=options.nCompsLocalPls;
    else
        nCompsLocalPls = [];
    end
    
    if isfield(options,'Amax')
        if length(options.Amax)==1
            Amax=repmat(options.Amax,1,length(X));
        else
            Amax=options.Amax;
        end
    else
        for i=1:length(X)
            Amax(i)=min([10 rank(X{i}.d)]);
        end
    end
    
    if isfield(options,'cvi')
        cvi=options.cvi;
    else cvi='r10';
    end
    
    if isfield(options,'nCompsForEachContribution')
        nCompsForEachContribution = options.nCompsForEachContribution;
    else
        nCompsForEachContribution = cell(1,length(blockCombinations));
    end
    
    if isfield(options,'autoselect')
        autoselect=options.autoselect;
    else
        autoselect=0;
    end
    
    
    
    if ischar(cvi)
        %set cv segments
        switch lower(cvi)
            case 'loo'
                cvi=1:nObj;
            case 'r10'
                nn=floor(nObj/10);
                rest=nObj-10*nn;
                cvi=[repmat(1:10,1,nn) 1:rest];
                cvi=cvi(randperm(nObj));
            case 'r20'
                nn=floor(nObj/20);
                rest=nObj-20*nn;
                cvi=[repmat(1:20,1,nn) 1:rest];
                cvi=cvi(randperm(nObj));
            otherwise
                if length(cvi)~=nObj
                    error('cvi segments does not match number of samples')
                end
        end
    end
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % initialise model
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    model.type = 'POSO-plsr';
    model.X=X; %save raw data
    model.Y=Y;
    model.options=options;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % preprocess X and Y
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    for i=1:nBlocks
        X{i}=mypreprocess(X{i},options.preprocX{i});
        X{i}=rmfield(X{i},'pp');
        totVarX{i}=sum(sum(X{i}.d.^2));
    end
    expVarX=cell(length(blockCombinations),nBlocks);
    
    Y=mypreprocess(Y,options.preprocY);
    Y=rmfield(Y,'pp');
    totVarY=sum(Y.d.^2);
    expVarY=cell(length(blockCombinations),1);
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Extract components from each of the blockCombinations
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    Yres=Y;
    Scores=cell(1,length(options.blockCombinations));
    Loadings=cell(1,length(options.blockCombinations));
    keep_local=0;
    
    
    
    for k=1:length(blockCombinations)
        idx=blockCombinations{k}; %which blocks to search for common components in
        
        if length(idx)>1 %more than one block
            
            if ~exist('localT','var') %local models have not yet been fitted
            local_models=fit_local_models(X,Yres,options);
            
            %extract scores from local models
            localT=cell(1,length(idx));
            localA=[];
            for i=1:nBlocks
                localA(i)=local_models{i}.Aopt;
                localT{i}=mat2saisir(local_models{i}.T(:,1:localA(i)));
            end
            nCompsLocalPls=localA;
            end
            %Canonical correlation analysis
            [A,r2,U,TC,forprediction{k}.predparamGCCA]=GCCA(localT(idx));
            
            
                
            %Select number of valid common components
            if isempty(nCompsForEachContribution{k})
                [nCompsForEachContribution{k},Scores{k},R2{k}]=select_common_components(TC,r2,totVarX,totVarY,X,Y,autoselect,idx);
            else
                Scores{k} = TC(:,nCompsForEachContribution{k});
                R2 = r2(:,nCompsForEachContribution{k});
            end
            
            
            if nCompsForEachContribution{k}~=0 %any components selected
                %calculate loadings, explained variances and orthogonalize (deflate)
                for i=1:length(nCompsForEachContribution{k}) % for each commonScores
                    for j=1:nBlocks
                        if sum(ismember(idx,j))>0 %calculate loadings only for the blocks in the current blockCombinations
                            Loadings{k}{j}(:,i)=X{j}.d'*Scores{k}(:,i);
                        end
                        
                        %all blocks are othogonolized, also those not in the
                        %current blockCombination
                        forprediction{k}.BorthT{j}(i,:)=Scores{k}(:,i)'*localT{j}.d;
                        forprediction{k}.BorthX{j}(i,:)=Scores{k}(:,i)'*X{j}.d;
                        localT{j}.d=localT{j}.d-Scores{k}(:,i)*forprediction{k}.BorthT{j}(i,:);
                        X{j}.d=X{j}.d-Scores{k}(:,i)*forprediction{k}.BorthX{j}(i,:);
                        expVarX{k,j}(end+1,1)=100-sum(sum(X{j}.d.^2))/totVarX{j}*100;
                    end
                    fittedY=Scores{k}(:,i)*Scores{k}(:,i)'*Yres.d;
                    
                end
                
                for j=1:length(idx)
                    forprediction{k}.W{j}=local_models{idx(j)}.W(:,1:localA(idx(j)));
                    forprediction{k}.P{j}=local_models{idx(j)}.P(:,1:localA(idx(j)));
                end
                
                
                %partial model
                allScores=[cell2mat(Scores)];
                beta=inv(allScores'*allScores)*allScores'*Y.d;
                
                if isequal(options.preprocY,'autoscale')
                    stdY=std(model.Y.d);
                else
                    stdY=ones(1,nY);
                end
                partialFittedY{k}=(allScores*beta).*repmat(stdY,nObj,1)+repmat(mean(model.Y.d),nObj,1);
                expVarY{k}=(1- sum((model.Y.d-partialFittedY{k}).^2)./sum((repmat(mean(model.Y.d),nObj,1)-model.Y.d).^2))*100;
              
                Yres.d=Yres.d-fittedY;
                partialResiduals{k}=Y.d-partialFittedY{k};
                
                
            else
                Scores{k}=[];
                Loadings{k}=[];
                
            end
            
            
            
        elseif length(idx)==1 %only one block
            
            if isempty(nCompsForEachContribution{k}) %need to select number of components
                
                m=mypls(X{idx},Yres,min([options.Amax(idx) rank(X{idx}.d)]),'mean center','mean center');
                m=mycrossval(m,cvi,0);
                
                if autoselect==0;
                    plsplots(m)
                    Aopt=input(['Block ' num2str(idx) ': How many uinque components?']);
                    close(gcf);
                else autoselect==1;
                    Aopt=m.Aopt;
                end
                nCompsForEachContribution{k} = Aopt;
            else %number of components is given
                m=mypls(X{idx},Yres,nCompsForEachContribution{k},'mean center','mean center');
            end
            
            forprediction{k}.P=m.P(:,1:nCompsForEachContribution{k});
            forprediction{k}.W=m.W(:,1:nCompsForEachContribution{k});
            forprediction{k}.V=forprediction{k}.W*inv(forprediction{k}.P'*forprediction{k}.W);
            if nCompsForEachContribution{k}>0; RegCoefs{k}=m.B(:,:,nCompsForEachContribution{k}); end
            
            if nCompsForEachContribution{k}>0 % calculate loadings and deflate X
                
                Scores{k}=m.T(:,1:nCompsForEachContribution{k});
                Loadings{k}=m.P(:,1:nCompsForEachContribution{k});
                
                
                for j=1:nBlocks % orthogonolize the remaining blocks
                    forprediction{k}.Borth{j}=inv(Scores{k}'*Scores{k})*Scores{k}'*X{j}.d;
                    X{j}.d = X{j}.d - Scores{k}*forprediction{k}.Borth{j};
                    expVarX{k,j}=100-sum(sum(X{j}.d.^2))/totVarX{j}*100;
                end
                
                %partial model
                allScores=[cell2mat(Scores)];
                beta=inv(allScores'*allScores)*allScores'*Y.d;
                
                if isequal(options.preprocY,'autoscale')
                    stdY=std(model.Y.d);
                else
                    stdY=ones(1,nY);
                end
                partialFittedY{k}=(allScores*beta).*repmat(stdY,nObj,1)+repmat(mean(model.Y.d),nObj,1);
                partialResiduals{k}=Y.d-partialFittedY{k};
                expVarY{k}=(1- sum((model.Y.d-partialFittedY{k}).^2)./sum((repmat(mean(model.Y.d),nObj,1)-model.Y.d).^2))*100;
                
                if nY>1
                    Yres.d = Yres.d - m.Ycal(:,:,nCompsForEachContribution{k}+1);
                else
                    Yres.d = Yres.d - m.Ycal(:,nCompsForEachContribution{k}+1);
                end
                
            end
            
        end
    end
    
    
    
    
    %final prediction model:
    allScores=[cell2mat(Scores)];
    beta=inv(allScores'*allScores)*allScores'*Y.d;
    
    if isequal(options.preprocY,'autoscale')
        stdY=std(model.Y.d);
    else
        stdY=ones(1,nY);
    end
    fittedY=(allScores*beta).*repmat(stdY,nObj,1)+repmat(mean(model.Y.d),nObj,1);
    
    
    %organize results
    
    commonIdx=[];
    uniqueIdx=[];
    for i=1:length(blockCombinations)
        if length(blockCombinations{i})>1 %common contribution
            if sum(nCompsForEachContribution{i})>0; commonIdx(end+1)=i; end
        else %unique contribution
            if sum(nCompsForEachContribution{i})>0; uniqueIdx(end+1)=i; end
        end
    end
    
    
    
    model.Scores = Scores([commonIdx uniqueIdx]);
    model.Loadings = Loadings([commonIdx  uniqueIdx]);
    model.fittedY = fittedY;
    model.beta=beta;
    if exist('RegCoefs','var'); model.RegCoefs=RegCoefs([commonIdx uniqueIdx]); end
    model.forprediction = forprediction;
    
    EVY=cell2mat(expVarY([commonIdx uniqueIdx]));
    if size(EVY,1)>1; EVY(2:end,:)=diff(EVY); end
    expVarY=mat2cell(EVY,ones(size(EVY,1),1),size(EVY,2));
    
    model.expVarX = expVarX([commonIdx uniqueIdx],:);
    model.expVarY = expVarY;
    model.RMSEC = sqrt(mean((model.fittedY-model.Y.d).^2));
    model.partialResiduals=partialResiduals([commonIdx uniqueIdx]);
    model.partialFittedY=partialFittedY([commonIdx uniqueIdx]);
    
    model.options.blockCombinations = blockCombinations([commonIdx uniqueIdx]);
    model.options.nCompsLocalPls = nCompsLocalPls;
    model.options.nCompsForEachContribution = nCompsForEachContribution([commonIdx uniqueIdx]);
    model.forprediction = forprediction([commonIdx uniqueIdx]);
    
    [anova_table,SSexpvar]=make_anova_table(model);
    
    model.ANOVA.table=anova_table;
    model.ANOVA.SSexpvar=SSexpvar;
    
    % check if there are common components or not (SO- or PO-PLS)
    for i=1:length(model.options.blockCombinations)
       contrlength(i)=length(model.options.blockCombinations{i}); 
    end
    if max(contrlength)==1; model.type='SO-PLS'; else model.type='PO_PLS'; end
    
    %calculating adjusted regression coefficients for SO-PLS
    if max(contrlength)==1 %SO-PLS
        if length(X)<4
%        model=adjust_RegCoefs_SOPLS(model);
        else
            disp('regression coefficients are not adjusted, adjustment works only for up to 3 blocks')
        end
        
    end
        
    
    
    
end


end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% local function used to fit separate pls models for each block
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function local_models=fit_local_models(X,Y,options)


for j=1:length(X) %number of blocks
    if options.nCompsLocalPls(j)==0
        %fit and cross-validate model
        local_models{j}=mypls(X{j},Y,min([options.Amax(j) rank(X{j}.d)]),options.preprocX{j},options.preprocY);
        local_models{j}=mycrossval(local_models{j},options.cvi,0);
        if options.autoselect==1 %automatic selection of components
            %select at least one component
            if local_models{j}.Aopt==0; local_models{j}.Aopt=1; end
        else
            plsplots(local_models{j});
            A=input(['Block ' num2str(j) ', Number of components: ']);
            if A==0; A=1; end
            local_models{j}.Aopt=A;
            close(gcf)
        end
        
        
    else     %fit model with predefined number of components
        
        
        local_models{j}=mypls(X{j},Y,options.nCompsLocalPls(j),options.preprocX{j},options.preprocY);
        local_models{j}.Aopt=options.nCompsLocalPls(j);
        
    end
    
end
end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% local function used to select number of common components
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [nComps,TC,R2]=select_common_components(TC,R2,totVarX,totVarY,X,Y,autoselect,idx)

[nObj,nY]=size(Y.d);
nBlocks=length(X);
nTC = length(R2);

resY=zeros(nObj,nTC,nY);
for j=1:size(TC,2)
    resY(:,j,1:nY)=TC(:,j)*TC(:,j)'*Y.d-Y.d; % dimension nObj x nTC x nY
end
resY = permute(resY,[2 3 1]); % nTC x nY x nObj
varianceY = sum(resY.^2,3); % nTC x nY

expVarY=(1-varianceY./repmat(totVarY,size(TC,2),1))*100;

expVarX = zeros(nTC,nBlocks);
for ii=1:nBlocks
    for jj=1:length(R2)
        expVarX(jj,ii)=sum(sum((TC(:,jj)*TC(:,jj)'*X{ii}.d).^2))/totVarX{ii}*100;
    end
end


%select number of common components based on R2, expVarY, expVarX
if autoselect == 0
    
    h=figure;
    set(h,'Name',['Common components, block ' num2str(idx)])
    if nTC==1
        hh=bar([expVarY expVarX; repmat(0,1,nBlocks+nY)]);
    else
        hh=bar([expVarY expVarX]);
    end
    set(hh(1:nY),'Facecolor','r')
    set(hh(1:nY),'Edgecolor','r')
    set(hh(nY+1:nY+nBlocks),'Facecolor','b')
    set(hh(nY+1:nY+nBlocks),'Edgecolor','b')
    
    hold on
    plot(1:length(R2),R2*100,'.','color','b')
    line([0 length(R2)+1],[90 90],'linestyle',':')
    line([0 length(R2)+1],[95 95],'linestyle',':')
    title('R^2, Explained variance in Y variables (red) and X blocks (blue)')
    
    nComps = input(['Common components, block ' num2str(idx) ', keep which components (vector)? ']);
    close(h)
elseif autoselect==1
    
        idx1=find(R2>0.90);
    idx2=find(max(expVarY')>5);
    idx3=find(min(expVarX(:,idx)')>5);
    nComps=intersect(intersect(idx1,idx2),idx3);
    
    if isempty(nComps); nComps=0; end
end

if sum(nComps)>0
    TC = TC(:,nComps);
    R2 = R2(:,nComps);
else
    TC=[];
    R2=[];
end

end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local function used to make ANOVA table
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [anova_table,SSexpvar]=make_anova_table(model)

[nObj,nY]=size(model.Y.d);
nBlocks=length(model.X);
SS=[];
rmsec=[];

tabtxt='Mean';
meanY=repmat(mean(model.Y.d),nObj,1);
SS(1,:)=diag((model.Y.d-meanY)'*(model.Y.d-meanY))';
rmsec(1,:)=sqrt(mean((model.Y.d-meanY).^2));

if isequal(model.options.preprocY,'autoscale')
    stdY=std(model.Y.d);
else
    stdY=ones(1,nY);
end

beta=model.beta;
beta_idx=0;
for i=1:length(model.options.blockCombinations)
    if length(model.options.blockCombinations{i})>1
        beta_idx=beta_idx+length(model.options.nCompsForEachContribution{i});
    else
        beta_idx=beta_idx+model.options.nCompsForEachContribution{i};
    end
    
    partial_beta=beta(1:beta_idx,:);
    
    partial_predictedY=(cell2mat(model.Scores(1:i))*partial_beta).*repmat(stdY,nObj,1)+repmat(mean(model.Y.d),nObj,1);
    
    SS(end+1,:)=diag((model.Y.d-partial_predictedY)'*(model.Y.d-partial_predictedY))';
    rmsec(end+1,:)=sqrt(mean((model.Y.d-partial_predictedY).^2));
    if length(model.options.blockCombinations{i})>1 %common components
        tabtxt=strvcat(tabtxt,['   Common (blocks ' num2str(model.options.blockCombinations{i}')' ')']);
    else %unique components
        tabtxt=strvcat(tabtxt,['   Unique block ' num2str(model.options.blockCombinations{i}) ]);
    end
    
end


tab1 = strvcat('Source',repmat('-',1,size(tabtxt,2)),tabtxt);
tab2 = repmat(' | ',size(tabtxt,1)+2,1);

%convert to explained variance
SSexpvar=(1-SS./repmat(SS(1,:),size(SS,1),1))*100;
SSexpvar=(round(SSexpvar*10))/10;

for i=1:nY
    SStxt=num2str(SSexpvar(:,i));
    rmsec_txt=num2str(rmsec(:,i));
    tab3 = strvcat('Exp Var(%) ',repmat('-',1,max(size(SStxt,2),11)),' ',SStxt(2:end,:));
    tab4 = strvcat(' rmsec ',repmat('-',1,max(size(rmsec_txt,2),11)),rmsec_txt);
    anova_table{i}=[tab1,tab2,tab3,tab4];
end
end